######################################################################
# regexps used by ZWiki

# in order to support more flexible numbers in wikinames, eg
# Issue0001, Issue0001SubjectHeading etc while not compromising
# simplicity, how about: "one or more digits is equivalent to a single
# capital letter" ?
# hmm getting a bit tricky

import re, string

# regular expressions (worth compiling these ?)

# URLs/URIs (use urllib/urlparse ?)
urlchars         = r'[A-Za-z0-9/:@_%~#=&\.\-\?\+\$,]+'
urlendchar       = r'[A-Za-z0-9/]'
url              = r'["=]?((about|gopher|http|https|ftp|mailto|file):%s)' % (urlchars)

# zwiki uses [...] to define arbitrary (non-WikiName) page names
# we'll work around stx footnote links if present
# \1 is expected to match what's inside the brackets
# also limit the characters to those zope will allow in an object id
# (should be kept in sync with OFS.ObjectManager.bad_id)
# / is not a valid id char but is handled differently by _wikilink_replace
idchars          = r'[a-zA-Z0-9-_~,.$\(\)# /]'
bracketedexpr    = r'\[(%s+?)\]' % (idchars)

# we'll do our own footnote handling
# real stx uses refchars = r'[0-9_%s-]' % (string.letters)
footnoteexpr     = r'(?sm)^\.\. \[(%s+?)\]' % (idchars)

# basic WikiName regexps
#
# orig
#wikiname1        = r'\b[A-Z]+[a-z]+[A-Z][A-Za-z]*[0-9]*'
#wikiname2        = r'\b[A-Z][A-Z]+[a-z][A-Za-z]*[0-9]*'
#
# use these for better i18n - you'll probably want to modify bad_id in
# zope's OFS/ObjectManager.py also
UC = string.uppercase                   
LC = string.lowercase
wikiname1        = r'\b[%s]+[%s]+[%s][%s]*[0-9]*' % (UC,LC,UC,UC+LC)
wikiname2        = r'\b[%s][%s]+[%s][%s]*[0-9]*' % (UC,UC,LC,UC+LC)
#
# a hack reported to work better for Russian-KO18 etc:
#import locale
#loc = locale.setlocale(locale.LC_ALL, 'ru_RU')
#wikiname1 = r'(?L)\b[A-Z\xE0-\xFF]+[a-z\xC0-\xDF]+[A-Z\xE0-\xFF][A-Za-z\xC0-\xFF]*[0-9]*'
#wikiname2 = r'(?L)\b[A-Z\xE0-\xFF][A-Z\xE0-\xFF]+[a-z\xC0-\xDF][A-Za-z\xC0-\xFF]*[0-9]*'
## in OFS/ObjectManager.py:
#bad_id=re.compile(r'[^a-zA-Z\xC0-\xFF0-9-_~,.$# ]').search

simplewikilink   = r'!?(%s|%s|%s)' % (wikiname1, wikiname2, url)
wikilink         = r'!?(%s|%s|%s|%s)' % (wikiname1,wikiname2,bracketedexpr,url)
interwikilink    = r'!?((?P<local>%s):(?P<remote>%s))' % \
                   (wikilink, urlchars+urlendchar)
remotewikiurl    = r'(?m)RemoteWikiURL[:\s]*([^\s]+)\s*$'
protected_line   = r'(?m)^!(.*)$'



# CMFWiki version:
#
#urlchars          = (r'[A-Za-z0-9/:@_%~#=&\.\-\?]+')
#url               = (r'["=]?((http|https|ftp|mailto|file|about):%s)'
#                     % (urlchars))
#urlexp            = re.compile(url)
## trying to co-exist with stx references:
#bracketedexpr     = r'\[([^\]0-9][^]]*)\]'
#bracketedexprexp  = re.compile(bracketedexpr)
#underlinedexpr    = r'_([^_]+)_'
#underlinedexprexp = re.compile(underlinedexpr)
#wikiname1         = r'\b[A-Z]+[a-z~]+[A-Z0-9][A-Z0-9a-z~]*'
#wikiname2         = r'\b[A-Z][A-Z0-9]+[a-z~][A-Z0-9a-z~]*'
#simplewikilinkexp = re.compile(r'!?(%s|%s)' % (wikiname1, wikiname2))
#
#wikiending        = r"[ \t\n\r\f\v:;.,<)!?']"
#urllinkending     = r'[^A-Za-z0-9/:@_%~\.\-\?]'
#wikilink          = (r'!?(%s%s|%s%s|%s|%s%s)'
#                    % (wikiname1,wikiending,
#                       wikiname2,wikiending,
#                       bracketedexpr,url,urllinkending))
#wikilinkexp       = re.compile(wikilink)
#wikilink_         = r'!?(%s|%s|%s|%s)' % \
#                     (wikiname1,wikiname2,bracketedexpr,url)
#interwikilinkexp  = re.compile(r'!?((?P<local>%s):(?P<remote>[\w]+))'
#                              % (wikilink_))
#remotewikiurlexp  = re.compile(r'(?m)RemoteWikiURL[:\s]*(.*)$')
#protected_lineexp = re.compile(r'(?m)^!(.*)$')
#
#antidecaptext     = '<!--antidecapitationkludge-->\n'
#antidecapexp      = re.compile(antidecaptext)
#
#commentsdelim = "<hr solid id=comments_below>"
#preexp = re.compile(r'<pre>')
#unpreexp = re.compile(r'</pre>')
#citedexp = re.compile(r'^\s*>')
## Match group 1 is citation prefix, group 2 is leading whitespace:
#cite_prefixexp = re.compile('([\s>]*>)?([\s]*)')
