# TrackerSupport mixin

import re
import DocumentTemplate
from OFS.DTMLDocument import DTMLDocument
from Regexps import url, bracketedexpr, wikiname1, wikiname2, simplewikilink, \
     wikilink, interwikilink, remotewikiurl, protected_line
from Utils import thunk_substituter, within_literal

class TrackerSupport:
    """
    This mix-in class adds some methods to ZWikiPage to support the
    zwiki tracker (see http://zwiki.org/ZwikiTracker)
    """
    def render_issuedtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        Render an "issue" page. This is an ordinary zwiki page with some
        extra properties appropriate for a bug/issue. These are displayed
        above the main text of the page in a sort of trouble ticket layout
        and can be edited.
        """
        issueform = DocumentTemplate.HTML ("""
<!-- start of issue property form -->
<form action="&dtml-page_url;/changeProperties" method="POST">
<table border="0" cellspacing="0" cellpadding="5"
<dtml-if "status == 'open'">
<dtml-if "severity == 'critical'">
bgcolor="#ff2222"
<dtml-elif "severity == 'serious'">
bgcolor="#ff6060"
<dtml-elif "severity == 'normal'">
bgcolor="#ffbbbb"
<dtml-elif "severity == 'minor'">
bgcolor="#ffdddd"
<dtml-elif "severity == 'wishlist'">
bgcolor="#ffe0e0"
<dtml-else>
bgcolor="#ffe0e0"
</dtml-if>
<dtml-elif "status == 'pending'">
bgcolor="#ffcc77"
<dtml-elif "status == 'closed'">
bgcolor="#e0f0e0"
<dtml-else>
bgcolor="#e0e0e0"
</dtml-if>
>
<tr><td>
**Issue title:**
<b><input type="text" name="title" value="<dtml-var title missing html_quote>" size="50" maxlength="100" style="font-weight:bold"></b>
<dtml-try>
<dtml-let
  highnumber="0#_.int(Catalog(page_type='issuedtml',sort_on='id')[-2].id[-4:])"
  thisnumber="_.int(id()[-4:])"
  previous="'IssueNo'+_.string.zfill(thisnumber-1,4)"
  next="'IssueNo'+_.string.zfill(thisnumber+1,4)"
>
<a href="&dtml-previous;">&lt;&lt;</a>&nbsp;<a href="&dtml-wiki_url;/ZwikiTracker?&dtml-QUERY_STRING;">^^</a>&nbsp;<a href="&dtml-next;">&gt;&gt;</a>
</dtml-let>
<dtml-except>
</dtml-try>
<br>
**Category:**
<select name="category">
<option <dtml-if "category=='don\\'t know'">selected</dtml-if>>don't know</option>
<option <dtml-if "category=='zwiki: installation'">selected</dtml-if>>zwiki: installation</option>
<option <dtml-if "category=='zwiki: general'">selected</dtml-if>>zwiki: general</option>
<option <dtml-if "category=='zwikidotorg template'">selected</dtml-if>>zwikidotorg template</option>
<option <dtml-if "category=='zwiki.org'">selected</dtml-if>>zwiki.org</option>
<option <dtml-if "category=='zwiki tracker'">selected</dtml-if>>zwiki tracker</option>
<option <dtml-if "category=='other'">selected</dtml-if>>other</option>
</select> 
**Severity:**
<select name="severity">
<option <dtml-if "severity=='critical'">selected</dtml-if>>critical</option>
<option <dtml-if "severity=='serious'">selected</dtml-if>>serious</option>
<option <dtml-if "severity=='normal'">selected</dtml-if>>normal</option>
<option <dtml-if "severity=='minor'">selected</dtml-if>>minor</option>
<option <dtml-if "severity=='wishlist'">selected</dtml-if>>wishlist</option>
</select> 
**Status:**
<select name="status">
<option <dtml-if "status=='open'">selected</dtml-if>>open</option>
<option <dtml-if "status=='pending'">selected</dtml-if>>pending</option>
<option <dtml-if "status=='closed'">selected</dtml-if>>closed</option>
</select> 
<input name="submit" type="submit" value="Change">
<br>
**Submitted by:**
<dtml-var "_.getattr(this(),'creator','') or '(unknown)'">
**at:**
<dtml-var creation_time missing=""><dtml-let
creation_time="_.getattr(this(),'creation_time','')"
creation="_.DateTime(_.getattr(this(),'creation_time','') or 1007105000)"
current="_.DateTime(ZopeTime().strftime('%Y/%m/%d %H:%M:%S'))"
elapsed="current-creation"
hourfactor="0.041666666666666664"
minutefactor="0.00069444444444444447"
secondsfactor="1.1574074074074073e-05"
days="_.int(_.math.floor(elapsed))"
weeks="days / 7"
months="days / 30"
years="days / 365"
hours="_.int(_.math.floor((elapsed-days)/hourfactor))"
minutes="_.int(_.math.floor((elapsed-days-hourfactor*hours)/minutefactor))"
seconds="_.int(_.round((elapsed-days-hourfactor*hours-minutefactor*minutes)/secondsfactor))"
>(<dtml-unless creation_time>unknown, ></dtml-unless><dtml-if years>
<dtml-var years> year<dtml-var "years > 1 and 's' or ''">
<dtml-elif months>
<dtml-var months> month<dtml-var "months > 1 and 's' or ''">
<dtml-elif weeks>
<dtml-var weeks> week<dtml-var "weeks > 1 and 's' or ''">
<dtml-elif days>
<dtml-var days> day<dtml-var "days > 1 and 's' or ''">
<dtml-elif hours>
<dtml-var hours> hour<dtml-var "hours > 1 and 's' or ''">
<dtml-elif minutes>
<dtml-var minutes> minute<dtml-var "minutes > 1 and 's' or ''">
<dtml-else>
<dtml-var seconds> second<dtml-var "seconds > 1 and 's' or ''">
</dtml-if> ago)
</dtml-let>
<br>
**Details & comments:**
</td></tr></table>
</form>
<!-- end of issue property form -->
""")                                    #'

        #dtml-evaluate both form and main page
        # not sure about this first one
        t = apply(issueform.__call__,(self, REQUEST),kw) + \
            apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)

        #wikilinks
        t = re.sub(protected_line, self._protect_line, t)
        t = self._structured_text(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        header = apply(self._renderHeaderOrFooter,
                       ('header',REQUEST,RESPONSE),kw)
        footer = apply(self._renderHeaderOrFooter,
                       ('footer',REQUEST,RESPONSE),kw)
        return header + t + footer

    def changeProperties(self, REQUEST=None, **kw):
        """
        identical to manage_changeProperties, except
        redirects back to the current page
        """
        if REQUEST is None:
            props={}
        else: props=REQUEST
        if kw:
            for name, value in kw.items():
                props[name]=value
        propdict=self.propdict()
        for name, value in props.items():
            if self.hasProperty(name):
                if not 'w' in propdict[name].get('mode', 'wd'):
                    raise 'BadRequest', '%s cannot be changed' % name
                self._updateProperty(name, value)

        self.reindex_object() # don't forget this
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.page_url())
            
    def category_index(self):
        """helper method to facilitate sorting catalog results"""
        try:
            return 1 + self.issue_categories.index(self.category)
        except (AttributeError,ValueError):
            return 0
        
    def severity_index(self):
        """helper method to facilitate sorting catalog results"""
        try:
            return 1 + self.issue_severities.index(self.severity)
        except (AttributeError,ValueError):
            return 0

    def status_index(self):
        """helper method to facilitate sorting catalog results"""
        try:
            return 1 + self.issue_statuses.index(self.status)
        except (AttributeError,ValueError):
            return 0

