# common initialization and support classes for zwiki unit tests

import string, re, os, sys

import unittest
# support older pyunit/python/zope
if not hasattr(unittest.TestCase,'assertEquals'):
    def assertEquals(self,first,second, msg=None):
        if first != second:
            raise self.AssertionError, (msg or '%s != %s' % (first, second))
    unittest.TestCase.assertEqual = \
        unittest.TestCase.assertEquals = assertEquals

# ensure "Products.ZWiki" will work in:
# normal zope, testrunner, direct invocation, python1.5, python2
# import App.FindHomes doesn't work because testrunner changes INSTANCE_HOME
import os
def pdir(path): return os.path.split(path)[0]
thisProductsDir = pdir(pdir(pdir(os.path.abspath(__file__))))
import Products
Products.__path__.insert(0,thisProductsDir)

# always needs to be imported early to set up Persistence.Persistent
import ZODB

# mock objects
import OFS.ObjectManager, AccessControl.User, DateTime
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from Products import ZWiki
from Products.ZWiki.ZWikiPage import ZWikiPage
from Products.ZWiki.SubscriberList import SubscriberList

class MockUser:
    def __init__(self,username='testuser'):
        self.username = username
        class aclusers: pass
        self.acl_users = aclusers()
        self.acl_users._nobody = 'nobody'
        
    def getUserName(self):
        return self.username


class MockRequest(HTTPRequest):
    # like makerequest without the app dependency
    def __init__(self):
        resp = HTTPResponse(stdout=sys.stdout)
        environ={}
        environ['SERVER_NAME']='foo'
        environ['SERVER_PORT']='80'
        environ['REQUEST_METHOD'] = 'GET'
        environ['SCRIPT_NAME']='/foo/test'
        HTTPRequest.__init__(self,None,environ,resp)

        
class MockZWikiPageInFolder(ZWikiPage,SubscriberList):

    def __init__(self, source_string='', mapping=None, __name__='TestPage',
                 **vars):
        #apply(ZWiki.ZWikiPage.ZWikiPage.__init__,
        #      (self,source_string,mapping,__name__),vars)
        apply(self.original.__init__,
              (self,source_string,mapping,__name__),vars)
        self.REQUEST = MockRequest()
        self.aq_inner = self
        #self.aq_parent = OFS.ObjectManager.ObjectManager()
        self.aq_parent = OFS.Folder.Folder()
        self.aq_parent.__class__.manage_addFolder = OFS.Folder.manage_addFolder
        self.aq_parent._setObject(self.getId(),self,set_owner=0)

    # work around acquisition shenanigans.. we seem to have an aq_parent
    # loop between page & objectmanager. 
    def getPhysicalPath(self): return ('',)

    def _checkPermission(self, permission, object):
        return 1

    ZopeTime = DateTime.DateTime

    def cb_isMoveable(self):
        return 1

# OMG! the hackery 8-)
# allows handleFileUpload to work
if not hasattr(MockZWikiPageInFolder,'original'):
    MockZWikiPageInFolder.original = ZWiki.ZWikiPage.ZWikiPage
    ZWiki.ZWikiPage.ZWikiPage = MockZWikiPageInFolder 

