# unit tests for ZWikiPage

from support import *
import OFS.Image
#from ZCatalog import ZCatalog,Vocabulary
#from ZCatalog.Catalog import Catalog,CatalogError


class TestZWikiPage(unittest.TestCase):
    def setUp(self):
        pass
        
    def tearDown(self):
        pass

    #def testZMIAddPageForm(self):
    #    ZWikiPage.manage_addZWikiPageForm(client=zc,
    #                                      REQUEST=zc.REQUEST)

    #def testZMIAddWebForm(self):
    #    ZWikiPage.manage_addZWikiPageForm(client=zc,
    #                                      REQUEST=zc.REQUEST)

    #def testZMIAddAndDelete(self):
    #    # sure this is the best way to get at manage_add ?
    #    zc.manage_addProduct['ZWiki'].manage_addZWikiPage('TestPageB', \
    #                                                       title='test page', \
    #                                                       file='blah blah')
    #    assert hasattr(zc,'TestPageB'), 'page not created'
    #    assert zc.TestPageB.text() == 'blah blah', 'wrong text'
    #    assert zc.TestPageB.page_type == 'structuredtext', \
    #           'wrong default page type'
    #    zc.manage_delObjects(['TestPageB'])
    #    assert not hasattr(zc,'TestPageB'), 'page not deleted'

    def testEdit(self):
        p = MockZWikiPageInFolder(__name__='TestPage')
        p.edit(text='something')
        self.assertEqual(p.read(),'something')
        p.edit(text='')
        self.assertEqual(p.read(),'')

    def testCreateFileOrImage(self):
        p = MockZWikiPageInFolder(__name__='TestPage')
        f = p.aq_parent
        import OFS.Image
        file = OFS.Image.Pdata('test file data')

        # our test page/folder initially has no uploads attr.
        self.assert_(not hasattr(p,'uploads'))

        # calling with an unnamed file should do nothing much
        self.assertEqual(p._createFileOrImage(file),(None, None, None))
        
        # ditto for a blank filename
        file.filename = ''
        self.assertEqual(p._createFileOrImage(file),(None, None, None))

        # here, a file object of unknown type should be created
        name = 'testfile'
        file.filename = name
        id, content_type,size = p._createFileOrImage(file)
        self.assertEqual(str(getattr(f,id)),'test file data')
        self.assertEqual(content_type ,'application/octet-stream')
        self.assertEqual(size,14)

        # a text file
        name = 'testfile.txt'
        file.filename = name
        id, content_type,size = p._createFileOrImage(file)
        self.assertEqual(str(f[id]),'test file data')
        self.assertEqual(content_type,'text/plain')

        # an image
        name = 'testfile.gif'
        file.filename = name
        id, content_type,size = p._createFileOrImage(file)
        # evaluating an Image gives its html tag
        self.assertEqual(str(f[id])[-47:] ,
                         '/testfile.gif" alt="testfile.gif" border="0" />')
        self.assertEqual(content_type,'image/gif')

    def testEditWithFileUpload(self):
        p = MockZWikiPageInFolder(__name__='TestPage')
        f = p.aq_parent

        # add a file to a page
        file = OFS.Image.Pdata('test file data')
        file.filename = 'edittestfile'
        p.REQUEST.file = file
        p.edit(REQUEST=p.REQUEST)

        # the new file should exist
        self.assert_(hasattr(f,'edittestfile'))
        # with the right data
        self.assertEqual(str(f['edittestfile']),'test file data')
        # and a link should have been added to the page
        self.assertEqual(p.read(),'\n\n<a href="edittestfile">edittestfile</a>\n')

        # a file with blank filename should be ignored
        p.REQUEST.file.filename = ''
        old = p.read()
        p.edit(REQUEST=p.REQUEST)
        self.assert_(p.read() == old)

        ## ditto, with an image
        p.edit(text='')
        file = OFS.Image.Pdata('test file data')
        file.filename = 'edittestimage.jpg'
        p.REQUEST.file = file
        p.edit(REQUEST=p.REQUEST)
        self.assert_(hasattr(f,'edittestimage.jpg'))
        self.assertEqual(f['edittestimage.jpg'].content_type,'image/jpeg')
        self.assertEqual(p.read(),'\n\n<img src="edittestimage.jpg">\n')

        # images should not be inlined if dontinline is set
        p.REQUEST.file.filename = 'edittestimage.png'
        p.REQUEST.dontinline = 1
        p.edit(REQUEST=p.REQUEST)
        self.assertEqual(p.read(),
          '\n\n<img src="edittestimage.jpg">\n\n\n<a href="edittestimage.png">edittestimage.png</a>\n')

    def test__handleDeleteMe_redirectToParent(self):
        p = MockZWikiPageInFolder()
        p.parents = ['chickens','dogs']
        p._recycle = lambda x: None
        req = MockRequest()
        class MyMockResponse:
            def redirect(self, url):
                self.redirectedto = url
        req.RESPONSE = MyMockResponse()

        # redirect to wiki url if no existing parents
        r = p._handleDeleteMe('DeleteMe',REQUEST=req)
        self.assertEqual(r,1)
        self.assert_(hasattr(req.RESPONSE,'redirectedto'))
        self.assertEqual(req.RESPONSE.redirectedto,
                         p.wiki_url()+'/')
        
        # redirect to first existing parent
        p.create(page='dogs')
        r = p._handleDeleteMe('DeleteMe',REQUEST=req)
        self.assertEqual(r,1)
        self.assert_(hasattr(req.RESPONSE,'redirectedto'))
        self.assertEqual(req.RESPONSE.redirectedto,
                         p.wiki_url()+'/'+'dogs')

    # can't test _recycle easily, because manage_pasteObject needs an app
    #def testDelete(self):
    #    p = MockZWikiPageInFolder(__name__='TestPage')
    #    f = p.aq_parent
    #    self.assert_(hasattr(f,'TestPage'))
    #    p._recycle()                     
    #    self.assert_(not hasattr(f,'TestPage'))
    #    self.assert_(hasattr(f,'recycle_bin'))
    #    self.assertEqual(f.recycle_bin.TestPage.read(),'\nDeleteMe')

    # same problem as above
    #def testEditDeleteMe(self):
    #    #When we see a first line beginning with "DeleteMe":
    #    #- move to recycle_bin
    #    #- redirect to first parent or default page
    #
    #    p = MockZWikiPageInFolder(__name__='TestPage')
    #    f = p.aq_parent
    #    self.assert_(hasattr(f,'TestPage'))
    #    self.assert_(not hasattr(f,'recycle_bin'))
    #
    #    #deleteme's not at the beginning shouldn't do anything
    #    p.edit(text=p.read()+'\nDeleteMe')
    #    self.assertEqual(p.read(),'\nDeleteMe')
    #
    #    #deleteme at the beginning will send it to recycle_bin
    #    p.edit(text='DeleteMe, with comments\n')
    #    self.assert_(not hasattr(f,'TestPage'))
    #    self.assert_(hasattr(f,'recycle_bin'))
    #    self.assertEqual(f.recycle_bin.TestPage.read(),'\nDeleteMe')
    
    def testEditLastEditorStamping(self):
        # Username stamping
        p = MockZWikiPageInFolder(__name__='TestPage')
        f = p.aq_parent
        p.last_editor = '-'

        # if no username available, IP address should be recorded
        p.REQUEST.set('REMOTE_ADDR', '1.2.3.4')
        p.append(text='.',REQUEST=p.REQUEST)
        self.assertEqual(p.last_editor,'1.2.3.4')
        self.assertEqual(p.last_editor_ip,'1.2.3.4')

        # use the zwiki_username cookie if available
        p.REQUEST.cookies['zwiki_username'] = 'cookiename'
        p.append(text='.',REQUEST=p.REQUEST)
        self.assertEqual(p.last_editor,'cookiename')
        self.assertEqual(p.last_editor_ip,'1.2.3.4')

        # if we are authenticated, use that by preference
        p.REQUEST.set('AUTHENTICATED_USER', MockUser('authusername'))
        p.append(text='.',REQUEST=p.REQUEST)
        self.assertEqual(p.last_editor,'authusername')
        self.assertEqual(p.last_editor_ip,'1.2.3.4')

        # don't record editor if nothing is actually changed
        p.REQUEST.set('AUTHENTICATED_USER', MockUser('differentusername'))
        p.REQUEST.set('REMOTE_ADDR', '5.6.7.8')
        p.edit(text=p.read(),REQUEST=p.REQUEST)
        self.assertEqual(p.last_editor,'authusername')
        self.assertEqual(p.last_editor_ip,'1.2.3.4')

    def test_create(self):
        p = MockZWikiPageInFolder()
        f = p.aq_parent
        
        # create a blank page
        p.create('TestPage1',text='')
        self.assert_(hasattr(f,'TestPage1'))
        self.assertEqual(f.TestPage1.text(),'')
        
        # create a classicwiki page with some text
        p.create('TestPage2',text='test page data',type='classicwiki')
        self.assert_(hasattr(f,'TestPage1'))
        self.assertEqual(f.TestPage2.read(),'test page data')
        self.assertEqual(f.TestPage2.page_type,'classicwiki')

    #def test_editform(self):
    #    p = MockZWikiPageInFolder()
    #    f = p.aq_parent

    def testCreateWithFileUpload(self):
        p = MockZWikiPageInFolder(__name__='TestPage1')
        f = p.aq_parent
        
        # upload a file while creating a page
        # this capability broke - fix if ever needed
        # and most of this is tested above I think
        file = OFS.Image.Pdata('test file data')
        file.filename = 'test_file'
        p.REQUEST.file = file
        p.create('TestPage3',text='test page data',REQUEST=p.REQUEST)
        # the new file should exist
        self.assert_(hasattr(f,'test_file'))
        # with the right data
        self.assertEqual(str(f['test_file']),'test file data')
        # and a link should have been added to the new wiki page
        self.assertEqual(f.TestPage3.read(),
          'test page data\n\n<a href="test_file">test_file</a>\n')

        # ditto with an image
        file.filename = 'test_image.gif'
        p.REQUEST.file = file
        f.TestPage1.create('TestPage4',text='test page data',REQUEST=p.REQUEST)
        self.assert_(hasattr(f,'test_image.gif'))
        self.assertEqual(f['test_image.gif'].content_type,'image/gif')
        self.assertEqual(f.TestPage4.read(),
                         'test page data\n\n<img src="test_image.gif">\n')

        # images should not be inlined if dontinline is set
        file.filename = 'test_image.JPG'
        p.REQUEST.dontinline = 1
        f.TestPage1.create('TestPage5',text='',REQUEST=p.REQUEST)
        self.assertEqual(f.TestPage5.read(),
                         '\n\n<a href="test_image.JPG">test_image.JPG</a>\n')

    def testIssueNo0112StxDoubleHtmlTag(self):
        p = MockZWikiPageInFolder()
        p.edit(type='structuredtext')
        text = p()
        self.assertEquals(
            len(re.findall(r'(?i)<html',text)),
            1)
        self.assertEquals(
            len(re.findall(r'(?i)<body',text)),
            1)

    def testCommentFormatting(self):
        USER = "me"
        TIME = "now"
        COMMENT = """\
short lines

aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa
aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa

long lines

bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb
bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb

quoted lines

> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc
> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc

long quoted lines

> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd
> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd
"""
        SHOULDWRITE = """\


<hr><b>me, now (via web):</b><br>
short lines

aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa
aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa

long lines

bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb
bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb

quoted lines

<br>><i> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc</i>
<br>><i> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc</i>

long quoted lines

<br>><i> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd</i>
<br>><i> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd</i>
"""
        SHOULDDIFF = """\


<hr><b>me, now (via web):</b><br>
short lines

aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa
aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa

long lines

bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb
bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb

quoted lines

<br>><i> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc</i>
<br>><i> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc</i>

long quoted lines

<br>><i> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd</i>
<br>><i> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dd</i>

"""
        SHOULDMAILOUT = """\
short lines

aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa
aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa aaaa

long lines

bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb
bbbb
bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb bbbb
bbbb

quoted lines

> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc
> cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc cccc

long quoted lines

> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd
dddd dd
> dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd dddd
dddd dd
"""
        COMMENT2 = "bah\n"
        SHOULDWRITE2 = "\n\nbah\n"
        SHOULDDIFF2 = "\n\nbah\n\n"
        SHOULDMAILOUT2 = "bah\n"

        from Products.ZWiki.Diff import ZWikiDiffMixin

        # test formatting at each stage, using a comment heading
        p = MockZWikiPageInFolder()
        oldtext = p.read()
        p.comment(text=COMMENT,username=USER,time=TIME,use_heading=1)
        self.assertEqual(p.read(), SHOULDWRITE)
        diff = ZWikiDiffMixin().textDiff(a=oldtext,b=p.read(),verbose=0)
        self.assertEqual(diff,SHOULDDIFF)
        self.assertEqual(p.formatMailout(diff), SHOULDMAILOUT)

        # without heading
        p.edit(text='')
        oldtext = p.read()
        p.comment(text=COMMENT2,username=USER,time=TIME,use_heading=0)
        self.assertEqual(p.read(), SHOULDWRITE2)
        diff = ZWikiDiffMixin().textDiff(a=oldtext,b=p.read(),verbose=0)
        self.assertEqual(diff,SHOULDDIFF2)
        self.assertEqual(p.formatMailout(diff), SHOULDMAILOUT2)


class TestZWikiPageCataloging(unittest.TestCase):
    # cf /usr/lib/zope/lib/python/Products/ZCatalog/tests/testCatalog.py
    #
    # we have various ZMI and zwiki methods
    # for adding, deleting, editing etc.
    # Easier to write doctests method by method ?
    #
    # these assume a default catalog is present and being used
    # beware of existing catalog data screwing things up, maybe
    # we can create our own (& set SITE_CATALOG) in zc

    def setUp(self):
        pass
    
    def tearDown(self):
        pass

#    def testOkWithoutCatalog(self):
#        pass

#    def testOkWithoutSiteCatalogProperty(self):
#        pass


    #def testCatalogDoctests(self):
    #XXX breaks on zwiki.org
    #>>> # the doctest setup/teardown creates & destroys TestPage -
    #>>> # use another in case of confusion
    #>>> # check everything is clean before we start
    #>>> zc.Catalog(id='CatalogTestPage')
    #[]
    #>>> zc.Catalog(PrincipiaSearchSource="blah")
    #[]
    #>>> zc.Catalog(PrincipiaSearchSource="blob")
    #[]
    #>>> # add a page via ZMI
    #>>> ZWiki.manage_addZWikiPage('CatalogTestPage',title='test page',\
    #                              file='blah')
    #''
    #>>> zc.Catalog(id='CatalogTestPage')[0][0]
    #'CatalogTestPage'
    #>>> zc.Catalog(PrincipiaSearchSource="blah")[0][0]
    #'CatalogTestPage'
    #>>> zc.Catalog(PrincipiaSearchSource="blob")
    #[]
    #>>> # edit a page via ZMI
    #>>> zc.CatalogTestPage.manage_edit('blob','')
    #>>> zc.Catalog(PrincipiaSearchSource="blah")
    #[]
    #>>> zc.Catalog(PrincipiaSearchSource="blob")[0][0]
    #'CatalogTestPage'
    #>>> # delete a page via ZMI
    #>>> zc.manage_delObjects(['CatalogTestPage'])
    #>>> zc.Catalog(id='CatalogTestPage')
    #[]


#XXX breaks on zwiki.org
#    def testAutoCatalogingWithZMIMethods(self):
#        """
#        shag this, let's go with doctest
#        it's easier to see what we're actually getting
#
#        XXX not currently doc-tested - see CatalogAwareness.py
#
#        even so there is a slight dilemma:
#        - use assert and get meaningful messages, or
#        - use the style below and always see what you got ?
#        nice to find an economical way to do both
#
#        now I tell you what might also be nice - the ability to inline
#        doctests with real code, and have them executed along with the
#        rest when python is run in debug mode.
#
#        >>> # the doctest setup/teardown creates & destroys TestPage -
#        >>> # use another in case of confusion
#        >>> # check everything is clean before we start
#        >>> zc.Catalog(id='CatalogTestPage')
#        []
#        >>> zc.Catalog(PrincipiaSearchSource="blah")
#        []
#        >>> zc.Catalog(PrincipiaSearchSource="blob")
#        []
#        >>> # add a page via ZMI
#        >>> ZWiki.manage_addZWikiPage('CatalogTestPage',title='test page',\
#                                      file='blah')
#        ''
#        >>> zc.Catalog(id='CatalogTestPage')[0][0]
#        'CatalogTestPage'
#        >>> zc.Catalog(PrincipiaSearchSource="blah")[0][0]
#        'CatalogTestPage'
#        >>> zc.Catalog(PrincipiaSearchSource="blob")
#        []
#        >>> # edit a page via ZMI
#        >>> zc.CatalogTestPage.manage_edit('blob','')
#        >>> zc.Catalog(PrincipiaSearchSource="blah")
#        []
#        >>> zc.Catalog(PrincipiaSearchSource="blob")[0][0]
#        'CatalogTestPage'
#        >>> # delete a page via ZMI
#        >>> zc.manage_delObjects(['CatalogTestPage'])
#        >>> zc.Catalog(id='CatalogTestPage')
#        []
#        >>> # who cleans this up if we don't ?
#        """
#        pass

#    def testAutoCatalogingWithZWikiMethods(self):
#        """
#        try this one in zunit
#        as noted above, these really expect to start with a clean Catalog
#        """
#        zc = self.ZopeContext
#        assert not zc.Catalog(id='CatalogTestPage'), \
#               'uncreated page found in the catalog'
#        
#        zc.manage_addProduct['ZWiki'].manage_addZWikiPage('CatalogTestPage1')
#        zc.CatalogTestPage1.edit(page='CatalogTestPage',text='bleh')
#        assert zc.Catalog(id='CatalogTestPage'), \
#               'page not found in catalog after creation by edit'
#        assert zc.Catalog(PrincipiaSearchSource="bleh"), \
#               'page text not found in catalog after creation by edit'
#        assert not zc.Catalog(PrincipiaSearchSource="blib"), \
#               'dummy text found in catalog after creation by edit'
#
#        zc.CatalogTestPage.append(text='blib')
#        assert zc.Catalog(PrincipiaSearchSource="blib"), \
#               'new text not found in catalog after append'
#
#        zc.CatalogTestPage.edit(text='DeleteMe')
#        assert not hasattr(zc,'CatalogTestPage'), 'ack! page not deleted'
#        assert not zc.Catalog(id='CatalogTestPage'), \
#               'page still found in catalog after DeleteMe edit'
        

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestZWikiPage))
    suite.addTest(unittest.makeSuite(TestZWikiPageCataloging))
    return suite

def main():
    unittest.TextTestRunner().run(test_suite())

if __name__ == '__main__':
    main()
